// Value.h: interface for TValue class
//

#pragma once

#include "Semaphore.h"

#undef DECL
#if defined(_AFXDLL) || defined(XPERT_EXE)
#ifdef UTILS_DLL
#define DECL _declspec(dllexport)
#else
#define DECL _declspec(dllimport)
#endif
#else
#define DECL 
#endif

class TIntRef;
class TDblRef;
class TCStrRef;


///////////////////////////////////////////////////////////////////////////
// TValue class
//
class DECL TValue
{
public:
   TValue()                               {DataType=0; RightDigits=-13;};
   virtual ~TValue()                      {};
   TValue(const TValue& o);
   TValue(TCHAR* Str)                     {TCriticalLock lock(cs); *this = Str; RightDigits=-13;};
   TValue(LPCTSTR Str)                    {TCriticalLock lock(cs); *this = Str; RightDigits=-13;};
   TValue(const CString Str)              {TCriticalLock lock(cs); *this = Str; RightDigits=-13;};
   TValue(int Num)                        {TCriticalLock lock(cs); *this = Num; RightDigits=-13;};
   TValue(bool Num)                       {TCriticalLock lock(cs); *this = Num; RightDigits=-13;};
   TValue(double Num)                     {TCriticalLock lock(cs); *this = Num; RightDigits=-13;};
   virtual TValue& operator=(const TValue& o);
   virtual TValue& operator=(TCHAR* Str)          {TCriticalLock lock(cs); DataType=STRING; String=Str; return *this;         };
   virtual TValue& operator=(LPCTSTR Str)         {TCriticalLock lock(cs); DataType=STRING; String=Str; return *this;         };
   virtual TValue& operator=(const CString Str)   {TCriticalLock lock(cs); DataType=STRING;  String=Str; return *this;        };
   virtual TValue& operator=(int Num)             {TCriticalLock lock(cs); DataType=INTEGER; Integer=Num; return *this;       };
   virtual TValue& operator=(bool Num)            {TCriticalLock lock(cs); DataType=INTEGER; Integer=(int) Num; return *this; };
   virtual TValue& operator=(double Num)          {TCriticalLock lock(cs); DataType=DOUBLE; Double=Num; return *this;         };
   // Update data but retain old type
   virtual void Update(const TValue& v);
//
   virtual TValue GetArray(int Index);
   virtual void SetArray(int Index, const TValue& Value);
   virtual int GetArraySize();
   virtual TValue operator[](int Index)           {TCriticalLock lock(cs); return GetArray(Index);                 };
   virtual operator int()                         {TCriticalLock lock(cs); MakeInt(); return Integer;              };
   virtual operator unsigned int()                {TCriticalLock lock(cs); MakeInt(); return Integer;              };
   virtual operator bool()                        {TCriticalLock lock(cs); MakeInt(); return Integer != 0;         };
   virtual operator CString()                     {TCriticalLock lock(cs); MakeStr(); return String;               };
   virtual operator double()                      {TCriticalLock lock(cs); MakeDbl(); return Double;               };
//
   virtual TValue& operator+=(TValue& v);
   virtual TValue& operator+=(double d);
   virtual TValue& operator+=(int n);
   virtual TValue& operator+=(CString s);
//
   virtual bool operator==(TValue& v);
   virtual bool operator==(double d)               {TCriticalLock lock(cs); MakeDbl(); return (Double == d);       };
   virtual bool operator==(int n)                  {TCriticalLock lock(cs); MakeInt(); return (Integer == n);      };
   virtual bool operator==(CString s)              {TCriticalLock lock(cs); MakeStr(); return (String == s);       };
//
   virtual bool operator!=(TValue& v);
   virtual bool operator!=(double d)               {TCriticalLock lock(cs); MakeDbl(); return (Double != d);       };
   virtual bool operator!=(int n)                  {TCriticalLock lock(cs); MakeInt(); return (Integer != n);      };
   virtual bool operator!=(CString s)              {TCriticalLock lock(cs); MakeStr(); return (String != s);       };
//
   virtual bool operator<(TValue& v);
   virtual bool operator<(double d)               {TCriticalLock lock(cs); MakeDbl(); return (Double < d);         };
   virtual bool operator<(int n)                  {TCriticalLock lock(cs); MakeInt(); return (Integer < n);        };
   virtual bool operator<(CString s)              {TCriticalLock lock(cs); MakeStr(); return (String < s);         };
//
   virtual bool operator>(TValue& v);
   virtual bool operator>(double d)               {TCriticalLock lock(cs); MakeDbl(); return (Double > d);         };
   virtual bool operator>(int n)                  {TCriticalLock lock(cs); MakeInt(); return (Integer > n);        };
   virtual bool operator>(CString s)              {TCriticalLock lock(cs); MakeStr(); return (String > s);         };
//
   virtual bool operator<=(TValue& v);
   virtual bool operator<=(double d)              {TCriticalLock lock(cs); MakeDbl(); return (Double <= d);        };
   virtual bool operator<=(int n)                 {TCriticalLock lock(cs); MakeInt(); return (Integer <= n);       };
   virtual bool operator<=(CString s)             {TCriticalLock lock(cs); MakeStr(); return (String <= s);        };
//
   virtual bool operator>=(TValue& v);
   virtual bool operator>=(double d)              {TCriticalLock lock(cs); MakeDbl(); return (Double >= d);        };
   virtual bool operator>=(int n)                 {TCriticalLock lock(cs); MakeInt(); return (Integer >= n);       };
   virtual bool operator>=(CString s)             {TCriticalLock lock(cs); MakeStr(); return (String >= s);        };
//
   virtual LPCTSTR AsLPCTSTR()                    {TCriticalLock lock(cs); MakeStr(); return (LPCTSTR) String;     };
   virtual const int AsInteger()                  {TCriticalLock lock(cs); MakeInt(); return Integer;              };
   virtual const CString& AsCString()             {TCriticalLock lock(cs); MakeStr(); return String;               };
   virtual const double AsDouble()                {TCriticalLock lock(cs); MakeDbl(); return Double;               };

   // These return non-const references, to support passing TValues by reference.
   // Note how the datatype is reset, ***which may cause a change to the base type.***
   virtual TIntRef AsIntRef();
   virtual TCStrRef AsCStrRef();
   virtual TDblRef AsDblRef();

   virtual bool IsInteger()                       {TCriticalLock lock(cs); return (DataType & IS_INT) != 0;        };
   virtual bool IsDouble()                        {TCriticalLock lock(cs); return (DataType & IS_DBL) != 0;        };
   virtual bool IsString()                        {TCriticalLock lock(cs); return (DataType & IS_STR) != 0;        };
   virtual bool IsNull()                          {TCriticalLock lock(cs); return DataType == 0;                   };
   virtual void SetRightDigits(int Digits);
   virtual int GetRightDigits()                   {TCriticalLock lock(cs); return RightDigits;                     };
   virtual void SetType(int Type)                 {TCriticalLock lock(cs); DataType = Type;                        };
   virtual int GetType()                          {TCriticalLock lock(cs); return DataType;                        };

   virtual void Lock()                            { cs.Lock();                    };
   virtual void Unlock()                          { cs.Unlock();                  };
   virtual bool LockTry()                         { return cs.LockTry();          };
protected:
   void MakeInt()                         {if ((DataType & CVT_INT)==0) {if (DataType & DOUBLE) Integer=(int) Double; else if (DataType & STRING) Integer=_ttoi(String); else Integer=0; DataType|=CVT_INT;}; };
   void MakeStr()                         {if ((DataType & CVT_STR)==0) {if (DataType & DOUBLE) {if (RightDigits < 0) String.Format(_T("%.*g"), -RightDigits, Double); else String.Format(_T("%1.*f"), RightDigits, Double);} else if (DataType & INTEGER) String.Format(_T("%d"), Integer); else String=_T(""); DataType|=CVT_STR;}}
   void MakeDbl()                         {if ((DataType & CVT_DBL)==0) {if (DataType & INTEGER) Double=Integer; else if (DataType & STRING) Double=_tcstod(String, NULL); else Double=0.0; DataType|=CVT_DBL;}; };
protected:
   enum TDataType {IS_INT=1, IS_DBL=2, IS_STR=4, CVT_INT=8, CVT_DBL=16, CVT_STR=32, INTEGER=IS_INT|CVT_INT, DOUBLE=IS_DBL|CVT_DBL, STRING=IS_STR|CVT_STR};
   int DataType;
   int Integer;
   double Double;
   CString String;
   int RightDigits;
   TCriticalSection cs;
};

/*
   The following classes (TIntRef, TDblRef, TCStrRef) exists to allow
   a reference to be returned to a TValue that is compatible with simple
   types but yet will cause the owning TValue to be updated in an atomic
   manner when the (typically temporary) reference object is destroyed.

   Functions do not need to be modified to accept these types - you can
   just keep using simple types.

   example:

      // Here's a simple function that takes an integer reference and zeros it:
      void ZeroInt(int& x)
      {
         x = 0;
      }

      // Let's call ZeroInt() passing it a TValue
      TValue v(2);
      ZeroInt(v.AsIntRef());
      ASSERT((int) v == 0);

   What happened here is that v.AsIntRef() created a temporary instance
   of TIntRef on the stack, and passed it to ZeroInt. Thanks to operator
   overloads it actually passes a reference to it's internal copy of
   the data. Hence, when ZeroInt sets x to 0, v does not change. 
   Only once the temporary TIntRef instance is destroyed does v 
   take on the new value (0 in this case).

*/

///////////////////////////////////////////////////////////////////////////
// TIntRef class
//
class TIntRef
{
public:
   TIntRef(TValue& v)           { pValue = &v; Data = (int) v;      };
   TIntRef(const TIntRef& r)    { pValue = r.pValue; Data = r.Data; };
   ~TIntRef()                   { *pValue = Data;                   };
   operator int&()              { return Data;                      };
   TIntRef& operator=(int data) { Data = data; return *this;        };
   int& AsRef()                 { return Data;                      };
protected:
   int Data;               // This must come first such that CString::Format and printf can work
   TValue* pValue;
};

///////////////////////////////////////////////////////////////////////////
// TDblRef class
//
class TDblRef
{
public:
   TDblRef(TValue& v)              { pValue = &v; Data = (double) v;   };
   TDblRef(const TDblRef& r)       { pValue = r.pValue; Data = r.Data; };
   ~TDblRef()                      { *pValue = Data;                   };
   operator double&()              { return Data;                      };
   TDblRef& operator=(double data) { Data = data; return *this;        };
   double& AsRef()                 { return Data;                      };
protected:
   double Data;               // This must come first such that CString::Format and printf can work
   TValue* pValue;
};

///////////////////////////////////////////////////////////////////////////
// TCStrRef class
//
class TCStrRef
{
public:
   TCStrRef(TValue& v)               { pValue = &v; Data = (CString) v;  };
   TCStrRef(const TCStrRef& r)       { pValue = r.pValue; Data = r.Data; };
   ~TCStrRef()                       { *pValue = Data;                   };
   operator CString&()               { return Data;                      };
   TCStrRef& operator=(CString data) { Data = data; return *this;        };
   CString& AsRef()                  { return Data;                      };
protected:
   CString Data;               // This must come first such that CString::Format and printf can work
   TValue* pValue;
};

